/* decode.c
 * Handlers for decoding a kexis stream into a wav stream.
 * Copyright (C) 2000 Wayde Milas (wmilas@rarcoa.com)
 * 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include "types.h"
#include "decode.h"
#include "bits.h"
#include "header.h"
#include "rice.h"
#include "predictor.h"
#include "kexis.h"

void decompress(KEXISHEADER *kexisHead, OPTIONSTRUCT *options)
{
	unsigned long totalBlocks;
	KEXISBLOCKSTRUCT kexisBlock;
	PCMBLOCKSTRUCT pcmBlock;
	time_t elapsedTime;
	PREDICTORTABLE *pred;

	options->inFileStream = fopen(options->inFileName, "r");
  if(options->inFileStream == NULL) {
    sprintf(options->errorString, "Can't open file: %s!", options->inFileName);
    handle_error(options);
  }

	// Suck in the KexisHeader
	parse_kexis_header(kexisHead, options);

	totalBlocks = kexisHead->dataLength;
	pcmBlock.pcmStreamLength = kexisHead->dataLength;
	totalBlocks = totalBlocks / 2;

	kexisBlock.data = malloc((sizeof(short)*UNITS));
	pcmBlock.data = malloc(sizeof(short) * UNITS);
	zero_predictor_table(&kexisBlock.predictor);
	kexisBlock.dataPosition = sizeof(KEXISHEADER);
	kexisBlock.blockPointer=0;
	kexisBlock.subBlockPointer=0;
	pcmBlock.size=0;
	kexisBlock.sumTotalSize =0;
	pred = &kexisBlock.predictor;
	
	while(kexisBlock.sumTotalSize < totalBlocks) {

		if(options->progress && options->dec_stdout==0
			&& (kexisBlock.sumTotalSize % 100000)==0) {
			printf("Progress:%6.2f%%   ",
				((float)kexisBlock.sumTotalSize/(float)totalBlocks)*100);
			time(&elapsedTime);
			printf("Time: %ld:%.2ld   ", (elapsedTime - options->time)/60,
				(elapsedTime - options->time)%60);
			if(options->encoderVersion == RICE_SINGLE_K)
				printf("K:%3d   Pd(mid,ave):%6d,%6d\r", pred->mid_K[0],
					pred->diffMid, pred->diffAve);
			else if(options->encoderVersion == RICE_DUAL_K) {
				printf("K(m,a):%3d,%3d   ", pred->mid_K[0], pred->ave_K[0]);
				printf("Pd(m,a):%6d,%6d\r", pred->diffMid, pred->diffAve);
		  }
			else if(options->encoderVersion == RICE_DUAL_K_BLOCK) {
				printf("K(m,a):%3d,%3d   ",
					block_ave_K(pred->mid_K, options->kHistorySize),
					block_ave_K(pred->ave_K, options->kHistorySize));
				printf("Pd(m,a):%6d,%6d\r", pred->diffMid, pred->diffAve);
			}
			fflush(stdout);
		}

		decode_kexisblock(&kexisBlock, &pcmBlock, options);
	}

	write_pcmblock(options, &kexisBlock, &pcmBlock, 0, 1);
	
	handle_verbose(options, &kexisBlock, &pcmBlock);
}

void decode_kexisblock(KEXISBLOCKSTRUCT *kexisBlock, PCMBLOCKSTRUCT *pcmBlock,
	OPTIONSTRUCT *options)
{
	long  diffMid, diffAve, mid, ave;
	short	right, left;

	rice_decode(kexisBlock, &diffMid, options, pcmBlock, MID);
	rice_decode(kexisBlock, &diffAve, options, pcmBlock, AVE);

	difference_calc(kexisBlock, options->predictorVersion, diffMid, diffAve,
		DECOMPRESS);

	mid = kexisBlock->predictor.diffMid;
	ave = kexisBlock->predictor.diffAve;

	// If we are gathering statistics, gather them :P
	if(options->verbose) 
		collect_prediction_stats(kexisBlock, diffMid, diffAve);

	right = ave - (mid/2);
	left = right + mid;

	if(PRED_DEBUG == 1 && kexisBlock->sumTotalSize < 100000) {
		printf("%ld:  %ld  %ld  %ld  %ld  %d  %d\n", kexisBlock->sumTotalSize,
	  diffMid, diffAve, mid, ave, left, right);
	 }

	kexisBlock->sumTotalSize +=2;
	
	write_pcmblock(options, kexisBlock, pcmBlock, left, 0);
	write_pcmblock(options, kexisBlock, pcmBlock, right, 0);
}

void write_pcmblock(OPTIONSTRUCT *options, KEXISBLOCKSTRUCT *kexisBlock,
	PCMBLOCKSTRUCT *pcmBlock, short data, int now)
{
	int length;
	WAVHEADER	wavHeader;

	if(options->outFileStream == NULL) {
		if(options->dec_stdout == 0) {
			options->outFileName = malloc(strlen(options->inFileName)+1);
			strncpy(options->outFileName, options->inFileName,
				strlen(options->inFileName)-3);
			strcat(options->outFileName, "wav\0");
		
			options->outFileStream = fopen(options->outFileName, "w");
			if(options->outFileStream == NULL) {
				sprintf(options->errorString, "Can't open file: %s!",
					options->outFileName);
				handle_error(options);
			}
		}	
		else {
			options->outFileStream = stdout;
		}
			
		create_wave_header(&wavHeader, pcmBlock->pcmStreamLength);
		length = fwrite(&wavHeader, sizeof(wavHeader), 1,
			options->outFileStream);
		if(length != 1) {
			sprintf(options->errorString, "Error writing Header to %s!",
				options->outFileName);
			handle_error(options);
		}
	}

	if(now == 0) {
		pcmBlock->data[pcmBlock->size] = data;
		pcmBlock->size++;
	}

	if((pcmBlock->size == UNITS) || (now==1)) {
		length = fwrite(pcmBlock->data, sizeof(short), pcmBlock->size,
			options->outFileStream);
		if(length != pcmBlock->size) {
			 sprintf(options->errorString, "Error writing to %s!",
			 options->outFileName);
			 handle_error(options);
		}
		pcmBlock->size = 0;
	}
}
